{ File    : $URL: file:///D:/SVN/svnDelphi/branches/A%20New%20Beginning/VISExt/servicing/src/VISExt.dpr $
  Author  : $Author: Deniz Oezmen $
  Changed : $Date: 2023-05-30 01:13:35 +0200 (Di, 30 Mai 2023) $
  Revision: $Rev: 659 $

  Visionaire Engine VIS extractor, originally based on Luigi Auriemma's
  visionaire.bms.
}
program VISExt;

{$APPTYPE CONSOLE}

{$IFNDEF VIS_VERSION_5}
  {$IFNDEF CONVERSION_RESTRICTED}
    {$R '..\res\main\version3.res' '..\res\main\version3.rc'}
  {$ELSE}
    {$R '..\res\restricted\version3.res' '..\res\restricted\version3.rc'}
  {$ENDIF}
{$ELSE}
  {$IFNDEF CONVERSION_RESTRICTED}
    {$R '..\res\main\version5.res' '..\res\main\version5.rc'}
  {$ELSE}
    {$R '..\res\restricted\version5.res' '..\res\restricted\version5.rc'}
  {$ENDIF}
{$ENDIF}

// set LARGEADDRESSAWARE flag
{$SetPEFlags $20}

uses
  KeyCfg    in 'backend\KeyCfg.pas',
  {$IFNDEF CONVERSION_RESTRICTED}
  CleanVBIN in 'convert\CleanVBIN.pas',
  {$ENDIF}
  CleanXML  in 'convert\CleanXML.pas',
  Compr     in 'convert\Compr.pas',
  ConvVBIN  in 'convert\ConvVBIN.pas',
  {$IFNDEF CONVERSION_RESTRICTED}
  ConvXML   in 'convert\ConvXML.pas',
  {$ENDIF}
  Crypto    in 'convert\Crypto.pas',
  {$IFNDEF VIS_VERSION_5}
  Names3    in 'convert\Names3.pas',
  {$ELSE}
  Names5    in 'convert\Names5.pas',
  {$ENDIF}
  VEDMap    in 'convert\VEDMap.pas',
  {$IFNDEF CONVERSION_RESTRICTED}
  VEDMapEx  in 'convert\VEDMapEx.pas',
  {$ENDIF}
  Shared    in 'shared\Shared.pas',
  VISFmt    in 'shared\VISFmt.pas',
  CmdLine   in 'utils\CmdLine.pas',
  ConOut    in 'utils\ConOut.pas',
  DCRC      in 'utils\DCRC.pas',
  MemSearch in 'utils\MemSearch.pas',
  ModInfo   in 'utils\ModInfo.pas',
  SdcpMD5   in 'utils\SdcpMD5.pas',
  Classes,
  StrUtils,
  SysUtils,
  IniFiles,
  Windows,
  ZLib,
  NativeXml;

var
  DetectedKey: AnsiString;

function ReadHeader(var VISFile: file; out VISHeader: TVISHeader): Boolean;
var
  BytesRead: Longint;
  NumFiles : Longword;
begin
  BlockRead(VISFile, VISHeader, SizeOf(VISHeader), BytesRead);

  NumFiles := VISHeader.NumFiles;

  BigEndian := True;
  SwitchEndian(VISHeader.NumFiles);

  if NumFiles < VISHeader.NumFiles then
  begin
    BigEndian := False;
    VISHeader.NumFiles := NumFiles
  end;

  Result := (SizeOf(VISHeader) = BytesRead) and (VISHeader.ID = VISHeaderID)
end;

function ConvertBinaryXMLToTextualXML(var BINBuffer: TBuffer;
  out XMLBuffer: TBuffer): Boolean;
begin
  // convert binary XML data to a structured XML document without performing
  // an identifier-mapping (i.e. all identifiers take on the form "T%d")
  Result := ConvertVBINToXML(BINBuffer, XMLBuffer, False)
end;

function ConvertBinaryXMLToVED(var BINBuffer: TBuffer;
  out VEDBuffer: TBuffer): Boolean;
{$IFNDEF CONVERSION_RESTRICTED}
var
  VEBBuffer: TBuffer;
{$ENDIF}
begin
  {$IFNDEF CONVERSION_RESTRICTED}
  // clean up path names, ...
  Result := CleanBinaryXML(BINBuffer, VEBBuffer);
  if not Result then
    Exit;

  // ... throw away binary buffer to conserve memory ...
  SetLength(BINBuffer, 0);

  // ... and convert binary XML data to a structured XML document (the VED) by
  // performing the identifier-mapping
  Result := ConvertVBINToXML(VEBBuffer, VEDBuffer, True)
  {$ELSE}
  // this code path should never be reached
  Result := False
  {$ENDIF}
end;

function ConvertBinaryXMLToVEB(var BINBuffer: TBuffer;
  BinaryXMLCompHeader: TBinaryXMLCompHeader; out VEBBuffer: TBuffer): Boolean;
{$IFNDEF CONVERSION_RESTRICTED}
var
  TempBuffer: TBuffer;
{$ENDIF}
begin
  {$IFNDEF CONVERSION_RESTRICTED}
  // clean up path names
  Result := CleanBinaryXML(BINBuffer, VEBBuffer);
  if not Result then
    Exit;

  // throw away binary buffer to conserve memory
  SetLength(BINBuffer, 0);

  // compress cleaned binary XML
  Result := Compress(VEBBuffer, TempBuffer);
  if not Result then
  begin
    SetLength(TempBuffer, 0);
    Exit
  end;

  // adapt header
  with BinaryXMLCompHeader do
  begin
    UncompSize := Length(VEBBuffer);
    CompSize   := Length(TempBuffer)
  end;

  // concatenate header and compressed data
  SetLength(VEBBuffer, Length(TempBuffer) + SizeOf(BinaryXMLCompHeader));
  Move(BinaryXMLCompHeader, VEBBuffer[0], SizeOf(BinaryXMLCompHeader));
  Move(TempBuffer[0], VEBBuffer[SizeOf(BinaryXMLCompHeader)],
    Length(TempBuffer));

  SetLength(TempBuffer, 0)
  {$ELSE}
  // this code path should never be reached
  Result := FALSE
  {$ENDIF}
end;

function ConvertTextualXMLToVED(var XMLBuffer: TBuffer;
  out VEDBuffer: TBuffer): Boolean;
{$IFNDEF CONVERSION_RESTRICTED}
var
  TempBuffer: TBuffer;
{$ENDIF}  
begin
  {$IFNDEF CONVERSION_RESTRICTED}
  // clean up path names, ...
  Result := CleanTextualXML(XMLBuffer, TempBuffer);
  if not Result then
    Exit;

  // ... throw away XML buffer to conserve memory ...
  SetLength(XMLBuffer, 0);

  // ...and perfom a node keyword conversion on the XML document (if necessary)
  Result := ConvertXMLToVED(TempBuffer, VEDBuffer)
  {$ELSE}
  // this code path should never be reached
  Result := False;
  {$ENDIF}
end;

function CleanTextualXMLOnly(var XMLBuffer: TBuffer;
  out CleanBuffer: TBuffer): Boolean;
begin
  // just clean up the archive designators
  Result := CleanTextualXML(XMLBuffer, CleanBuffer)
end;

function GuessWebP(Buffer: TBuffer): Boolean;
var
  ID: AnsiString;
begin
  Result := False;
  if Length(Buffer) < 16 then
    Exit;

  // We just try and see whether we can find the chunk names and identifiers in
  // the right places ...
  SetLength(ID, 4);
  Move(Buffer[0], ID[1], Length(ID));
  if ID <> 'RIFF' then
    Exit;

  SetLength(ID, 8);
  Move(Buffer[8], ID[1], Length(ID));
  if (ID <> 'WEBPVP8 ') and (ID <> 'WEBPVP8X') and (ID <> 'WEBPVP8L') then
    Exit;

  Result := True
end;

function DetectGameKey(var VISFile: file; const VISHeader: TVISHeader;
  const Dir: TBuffer; out DirEntries: TDirEntries): Boolean;
var
  i, j     : Longint;
  DirHeader: TDirHeader;
  DirFooter: TDirFooter;
  TempKey  : AnsiString;
  SizeKey  : Longword;

  function DecryptAndCheck: Boolean;
  var
    i: Longint;
  begin
    Result := False;

    // Decrypt directory table with candidate key.
    Decrypt(Dir, 0, Length(Dir), DetectedKey);

    // Get header and footer identifiers ...
    Move(Dir[0], DirHeader, SizeOf(DirHeader));
    Move(Dir[Length(Dir) - SizeOf(DirFooter)], DirFooter, SizeOf(DirFooter));

    // ... and check for validity. If this passes, do a more thorough check of
    // the directory structure.
    if (DirHeader.ID = DirHeaderID) and (DirFooter.ID = DirFooterID) then
    begin
      // copy directory data to internal structures
      SetLength(DirEntries, VISHeader.NumFiles);
      Move(Dir[SizeOf(TDirHeader)], DirEntries[0],
        VISHeader.NumFiles * SizeOf(TDirEntry));

      for i := 0 to High(DirEntries) do
        with DirEntries[i] do
        begin
          SwitchEndian(Offset);
          SwitchEndian(CompSize);
          SwitchEndian(UncompSize);
          SwitchEndian(Flags)
        end;

      // Do some checks via the correlation of offset and size.
      Result := (DirEntries[0].Offset = 0);

      for i := 1 to High(DirEntries) do
        with DirEntries[i] do
          Result := Result and ((DirEntries[i - 1].Offset) +
            DirEntries[i - 1].CompSize = DirEntries[i].Offset)
    end;

    // wrong key: undo previous decryption
    if not Result then
      Decrypt(Dir, 0, Length(Dir), DetectedKey)
  end;

begin
  Result := False;

  // If an override key has been specified, then that one takes precedence.
  if ParamRec.OverrideKey <> '' then
  begin
    WriteLn('  Info: Using override key ' + ParamRec.OverrideKey);
    DetectedKey := ParamRec.OverrideKey;

    if DecryptAndCheck then
      Result := True
    else
      WriteLn('  Info: Override key is invalid.')
  end;

  // If the override key has not been specified or is invalid, try to detect the
  // key using the list of known game keys.
  if not Result then
    for i := 0 to High(GameConfigs) do
    begin
      DetectedKey := GameConfigs[i].Key;
      if DecryptAndCheck then
      begin
        WriteLn('  Info: Assuming ' + GameConfigs[i].Name + ' data file.');
        Result := True;
        Break
      end
    end;

  // If no known key matches and brute force is allowed, let's try to find the
  // key now. Most bytes can be diretly determined by using some assumptions.
  // Some of these only work, however, if the archive contains more than one
  // file.
  //
  // Note: This logic is quite messy, but should get the job done ...
  if not Result and ParamRec.Force then
  begin
    WriteLn('  Info: Trying to find unknown key.');

    // First set the parts of the key which can be directly assumed: The first
    // three bytes of the encrypted directory map to the plain text "HDR".
    // Following these is a doubleword representing the relative offset of the
    // first file -- which is always zero when decrypted. The last key byte
    // encrypts the highest byte of the flag entry, which can assumed to be zero
    // as well.
    //
    // The fields for the compressed size of the first file entry are identical
    // to the relative offset of the second file entry, should it exist. Since
    // the key part for the offset is known (see above), we can directly decrypt
    // this information as well, leaving only four bytes unknown.
    //
    // In case there is only one file entry within the archive, we can directly
    // compute its compressed size, since the sizes of all meta structures are
    // known.
    //
    // Note: We explicitly set the bytes 12 through 15 to zero here, since we
    // need the corresponding encrypted bytes in an unmodified state later on.
    SetLength(TempKey, 16);
    TempKey[ 1] := Chr(Dir[ 0] xor Ord(DirHeaderId[1]));
    TempKey[ 2] := Chr(Dir[ 1] xor Ord(DirHeaderId[2]));
    TempKey[ 3] := Chr(Dir[ 2] xor Ord(DirHeaderId[3]));
    TempKey[ 4] := Chr(Dir[ 3]);
    TempKey[ 5] := Chr(Dir[ 4]);
    TempKey[ 6] := Chr(Dir[ 5]);
    TempKey[ 7] := Chr(Dir[ 6]);

    if VISHeader.NumFiles > 1 then
    begin
      // We have at least two files: Use the offset of the second file to
      // get information on the partial key encrypting the compressed size.
      TempKey[ 8] := Chr(Dir[ 7] xor Dir[19] xor Ord(TempKey[4]));
      TempKey[ 9] := Chr(Dir[ 8] xor Dir[20] xor Ord(TempKey[5]));
      TempKey[10] := Chr(Dir[ 9] xor Dir[21] xor Ord(TempKey[6]));
      TempKey[11] := Chr(Dir[10] xor Dir[22] xor Ord(TempKey[7]));
    end
    else
    begin
      // We only have one file: Compute the compressed size of that file
      // directly and derive the partial key that way.
      SizeKey := FileSize(VISFile) - SizeOf(TVISHeader) - SizeOf(TDirHeader)
        - SizeOf(TDirFooter) - SizeOf(TDirEntry);
      SwitchEndian(SizeKey);
      Move(SizeKey, TempKey[8], 4);

      TempKey[ 8] := Chr(Dir[ 7] xor Ord(TempKey[ 8]));
      TempKey[ 9] := Chr(Dir[ 8] xor Ord(TempKey[ 9]));
      TempKey[10] := Chr(Dir[ 9] xor Ord(TempKey[10]));
      TempKey[11] := Chr(Dir[10] xor Ord(TempKey[11]))
    end;

    TempKey[12] := #000;
    TempKey[13] := #000;
    TempKey[14] := #000;
    TempKey[15] := #000;
    TempKey[16] := Chr(Dir[15]);

    // Decrypt as much information as possible up to now. If DecryptAndCheck
    // returns a failure, one of our basic assumptions above is wrong and we
    // cannot continue anyway.
    DetectedKey := TempKey;
    if DecryptAndCheck then
    begin
      // The remaining key bytes basically encrypt the uncompressed size of the
      // file entry. We will exploit the fact that usually only few files within
      // an archive are actually compressed, i.e. the values for compressed and
      // uncompressed size are the same.
      //
      // Thus, for each file entry in the archive, we choose our candidate key
      // based on this assumption. We consider the partial key to be valid when
      // it only consists of lower case hexadecimal values.
      //
      // (Note: If no valid key can be derived this way, we might assume that
      // all files within the archive are compressed. In that case, we could
      // take one of them, use a "large enough" output buffer, decompress the
      // file and see how big it gets. This would give us the remaining key
      // bytes as well. Alas, this approach is not implemented here.)  
      for i := 0 to High(DirEntries) do
      begin
        // Note: The decryption routine has possibly switched the endianness
        // of the size fields, so we have to undo this to get the
        // correct key order.
        SizeKey := DirEntries[i].CompSize xor DirEntries[i].UncompSize;
        SwitchEndian(SizeKey);
        Move(SizeKey, TempKey[12], 4);

        if (TempKey[12] in ['0'..'9', 'a'..'f']) and
           (TempKey[13] in ['0'..'9', 'a'..'f']) and
           (TempKey[14] in ['0'..'9', 'a'..'f']) and
           (TempKey[15] in ['0'..'9', 'a'..'f']) then
        begin
          // We seem to have found a winner. If so, we have to finish the
          // decryption now, since the uncompressed sizes are still encrypted.
          SwitchEndian(SizeKey);
          for j := 0 to High(DirEntries) do
            DirEntries[j].UncompSize := DirEntries[j].UncompSize xor SizeKey;

          DetectedKey := TempKey;
          Result := True;
          Break
        end
      end
    end;

    if not Result then
      WriteLn('  Info: No matching key found.')
    else
      WriteLn('  Info: Found key ' + DetectedKey)
  end
end;

procedure ExtractFiles(var VISFile: file; const VISName: AnsiString;
  const VISHeader: TVISHeader);
var
  i                  : Longint;
  BaseOfs,
  BufPos,
  UncompBufPos       : Longword;
  Dir                : TBuffer;
  DirEntries         : TDirEntries;
  XMLCompBlockHeader : TXMLCompBlockHeader;
  BinaryXMLCompHeader: TBinaryXMLCompHeader;
  RealName,
  OutName            : UTF8String;
  Buffer,
  UncompBuf,
  ConvBuf            : TBuffer;
  OutFile            : file;
begin
  // read directory table
  SetLength(Dir, SizeOf(TDirHeader) + VISHeader.NumFiles * SizeOf(TDirEntry) +
    SizeOf(TDirFooter));
  {$I-}
  BlockRead(VISFile, Dir[0], Length(Dir));
  {$I+}
  if IOResult <> 0 then
  begin
    WriteLn('  error reading structures');
    SetLength(Dir, 0);
    Exit
  end;

  if VISHeader.NumFiles = 0 then
  begin
    WriteLn('  archive is empty');
    SetLength(Dir, 0);
    Exit
  end;

  BaseOfs := FilePos(VISFile);

  // Try to detect the game key using the directory table identifiers as
  // markers. If successful, the result is the decrypted directory structure.
  if not DetectGameKey(VISFile, VISHeader, Dir, DirEntries) then
  begin
    WriteLn('  error detecting key');
    SetLength(Dir, 0);
    Exit
  end;

  if ParamRec.KeyOnly then
  begin
    SetLength(Dir, 0);
    Exit
  end;

  SetLength(Dir, 0);

  // process all files
  for i := 0 to High(DirEntries) do
    with DirEntries[i] do
    begin
      // Special case for a binary main XML file: This type of XML does not have
      // its own flag -- it is probably recognized for having the right magic
      // indentifier. It should, however, usually be the first (for savegames)
      // or last file (for game archives) in the archive. So in case we're
      // currently looking at the archive's first or last file, we'll check
      // whether this one might be the main XML.
      if ((i = 0) or (i = High(DirEntries))) and
        (CompSize >= SizeOf(BinaryXMLCompHeader)) then
      begin
        // The file is big enough to contain the binary XML compression header.
        // We'll read the first portion and check whether this might actually
        // hold true.
        {$I-}
        Seek(VISFile, BaseOfs + Offset);
        BlockRead(VISFile, BinaryXMLCompHeader, SizeOf(BinaryXMLCompHeader));
        {$I+}
        if IOResult <> 0 then
        begin
          WriteLn('    read error');
          Continue
        end;

        // If the header identifies as a binary XML, we set our "virtual" flag
        // for this file. The special case then ends here and the rest of the
        // handling is done below.
        if BinaryXMLCompHeader.ID = BinaryXMLID then
          Flags := Flags or Flag_BinaryMainXML
      end;

      // determine a name for this file
      GetName(VISName, DirEntries[i], i, RealName, OutName);

      // if a name table is about to be generated, we're only interested in the
      // main XML file
      if ParamRec.GenNames and ((Flags and Flag_AnyMainXML) = 0) then
        Continue;

      WriteLn('  ', Utf8ToAnsi(OutName));

      // read file data
      SetLength(Buffer, CompSize);
      {$I-}
      Seek(VISFile, BaseOfs + Offset);
      BlockRead(VISFile, Buffer[0], Length(Buffer));
      {$I+}
      if IOResult <> 0 then
      begin
        WriteLn('    read error');
        Continue
      end;

      // Correction for image files if we are not using a name file: The default
      // extension ".png" might be wrong: Let's try and see whether we can
      // identify a WebP file instead.
      if (not ParamRec.UseNames or (ParamRec.UseNames and
        ParamRec.UseDefaultName) or (ParamRec.UseNames and (RealName = ''))) and
        (AnsiLowerCase(ExtractFileExt(OutName)) = '.png') then
        if GuessWebP(Buffer) then
        begin
          OutName := ChangeFileExt(OutName, '.webp');
          WriteLn('    --> guessing ', Utf8ToAnsi(OutName))
        end;

      // Special case: If the "new" main XML file is found compressed, we have
      // to do a block-wise decompression (and decryption, if necessary).
      if ((Flags and Flag_MainXML2) <> 0) and (CompSize <> UncompSize) then
      begin
        BufPos := 0;
        UncompBufPos := 0;

        SetLength(UncompBuf, UncompSize);

        while Longint(BufPos + SizeOf(XMLCompBlockHeader)) <= High(Buffer) do
        begin
          // Get the compressed and uncompressed sizes of the next block.
          Move(Buffer[BufPos], XMLCompBlockHeader, SizeOf(XMLCompBlockHeader));
          Inc(BufPos, SizeOf(XMLCompBlockHeader));

          // The sizes are stored as big endian values, so switch them.
          with XMLCompBlockHeader do
          begin
            SwitchEndian(UncompSize, True);
            SwitchEndian(CompSize, True)
          end;

          // Check for the special marker designating the chain end.
          if XMLCompBlockHeader.CompSize = CompChainEnd then
            Break;

          // If the file is encrypted, we have to decrypt the block content
          // first.
          if (Flags and Flag_CryptDefault) <> 0 then
            Decrypt(Buffer, BufPos, XMLCompBlockHeader.CompSize, DetectedKey);

          // Now decompress the block content into the output buffer.
          if not Decompress(Buffer, BufPos, XMLCompBlockHeader.CompSize,
            UncompBuf, UncompBufPos, XMLCompBlockHeader.UncompSize) then
          begin
            WriteLn('    decompression error');
            Break
          end;

          BufPos := BufPos + XMLCompBlockHeader.CompSize;
          UncompBufPos := UncompBufPos + XMLCompBlockHeader.UncompSize
        end;

        SetLength(Buffer, 0);
        Buffer := UncompBuf
      end
      else if ((Flags and Flag_BinaryMainXML) <> 0) and
        (ParamRec.BINtoXML or ParamRec.BINtoVED or ParamRec.BINtoVEB or
        ParamRec.GenNames) then
      begin
        // Another special case for binary XML files: This type of file has to
        // be decompressed en bloc. However, we only do this when we also want
        // to convert it to a different representation or generate a name file,
        // since its not of much use else.
        BufPos := 0;

        // Get the compressed and uncompressed sized of the XML file.
        Move(Buffer[BufPos], BinaryXMLCompHeader, SizeOf(BinaryXMLCompHeader));
        Inc(BufPos, SizeOf(BinaryXMLCompHeader));

        // Check whether the binary XML is actually compressed. If we can assume
        // it is not, we'll just remove the compression header.
        if BinaryXMLCompHeader.CompSize <> BinaryXMLCompHeader.UncompSize then
        begin
          // Now decompress the block content into the output buffer.
          SetLength(UncompBuf, BinaryXMLCompHeader.UncompSize);
          if Decompress(Buffer, BufPos, BinaryXMLCompHeader.CompSize, UncompBuf,
            0, BinaryXMLCompHeader.UncompSize)
          then
          begin
            SetLength(Buffer, 0);
            Buffer := UncompBuf
          end
          else
            WriteLn('    decompression error')
        end
        else
          Move(Buffer[BufPos], Buffer[0], Longword(Length(Buffer)) - BufPos)
      end
      else
      begin
        // regular cases for the "old" main XML file and all other types of
        // files

        // decrypt the whole file if necessary
        if (Flags and Flag_CryptDefault) <> 0 then
          Decrypt(Buffer, 0, Length(Buffer), DetectedKey);

        // correct the image header if necessary
        if ((Flags and Flag_CryptImage) <> 0) and not ParamRec.NoImgDecrypt then
        begin
          // WebP
          if (AnsiLowerCase(ExtractFileExt(OutName)) = '.webp') then
          begin
            if ParamRec.UseNames and (RealName <> '') then
              DecryptWebP(Buffer, RealName)
            else if RealName = '' then
              WriteLn('    warning: no name found in name file; WebP decrypti' +
                'on not supported')
            else
              WriteLn('    warning: WebP decryption not supported without nam' +
                'e file')
          end
          else
          begin
            // PNG
            if ParamRec.UseNames then
            begin
              if not DecryptPNG(Buffer, RealName) then
                DecryptPNGForce(Buffer)
            end
            else
              DecryptPNGForce(Buffer)
          end
        end;

        // decompress if necessary
        if CompSize <> UncompSize then
        begin
          SetLength(UncompBuf, UncompSize);
          if Decompress(Buffer, 0, Length(Buffer), UncompBuf, 0,
            Length(UncompBuf)) then
          begin
            SetLength(Buffer, 0);
            Buffer := UncompBuf
          end
          else
            WriteLn('    decompression error')
        end
      end;

      // If the name file is to be generated, use the first XML file for this
      // task.
      if ParamRec.GenNames and ((Flags and Flag_AnyMainXML) <> 0) then
      begin
        WriteLn('    --> generating ', ParamRec.NameSpec);

        if (Flags and Flag_TextMainXML) <> 0 then
          GenerateNames(Buffer, False)
        else
          GenerateNames(Buffer, True);

        Break;
      end;

      // If the binary XML is to be converted to a textual XML format, try doing
      // so now.
      if ((Flags and Flag_BinaryMainXML) <> 0) and ParamRec.BINtoXML then
      begin
        // Switch type and file name for further processing.
        Flags := (Flags and not Flag_BinaryMainXML) or Flag_MainXML1;
        GetName(VISName, DirEntries[i], i, RealName, OutName);

        WriteLn('    --> converting to ', Utf8ToAnsi(OutName));
        if ConvertBinaryXMLToTextualXML(Buffer, ConvBuf) then
        begin
          SetLength(Buffer, 0);
          Buffer := ConvBuf
        end
        else
          WriteLn('    conversion error')
      end
      // If the binary XML is to be converted to the editable VED format, try
      // doing so now.
      else if ((Flags and Flag_BinaryMainXML) <> 0) and ParamRec.BINtoVED then
      begin
        // Switch type and file name for further processing.
        Flags := (Flags and not Flag_BinaryMainXML) or Flag_VEDMainXML;
        GetName(VISName, DirEntries[i], i, RealName, OutName);

        WriteLn('    --> converting to ', Utf8ToAnsi(OutName));
        if ConvertBinaryXMLToVED(Buffer, ConvBuf) then
        begin
          SetLength(Buffer, 0);
          Buffer := ConvBuf
        end
        else
          WriteLn('    conversion error')
      end
      // If the binary XML is to be converted to the VEB format (which
      // corresponds to "simply" cleaning the path names in the binary format
      // and recompressing the file), try doing so now.
      else if ((Flags and Flag_BinaryMainXML) <> 0) and ParamRec.BINtoVEB then
      begin
        // Switch type and file name for further processing.
        Flags := (Flags and not Flag_BinaryMainXML) or Flag_VEBMainXML;
        GetName(VISName, DirEntries[i], i, RealName, OutName);

        WriteLn('    --> converting to ', Utf8ToAnsi(OutName));
        if ConvertBinaryXMLToVEB(Buffer, BinaryXMLCompHeader, ConvBuf) then
        begin
          SetLength(Buffer, 0);
          Buffer := ConvBuf
        end
        else
          WriteLn('    conversion error')
      end;

      // If the XML is supposed to be cleaned of archive designators, do so now.
      if ParamRec.CleanXML and ((Flags and Flag_TextMainXML) <> 0) then
      begin
        if CleanTextualXMLOnly(Buffer, ConvBuf) then
        begin
          SetLength(Buffer, 0);
          Buffer := ConvBuf
        end
        else
          WriteLn('    cleanup error')
      end
      // If the XML is to be converted to the editable VED format, do so now.
      else if ParamRec.XMLtoVED and ((Flags and Flag_TextMainXML) <> 0) then
      begin
        // Switch type and file name for further processing.
        Flags := (Flags and not Flag_TextMainXML) or Flag_VEDMainXML;
        GetName(VISName, DirEntries[i], i, RealName, OutName);

        WriteLn('    --> converting to ', Utf8ToAnsi(OutName));
        if ConvertTextualXMLToVED(Buffer, ConvBuf) then
        begin
          SetLength(Buffer, 0);
          Buffer := ConvBuf
        end
        else
          WriteLn('    conversion error')
      end;

      // open output file and write final data
      if not OpenOutFile(OutFile, OutName) then
      begin
        WriteLn('    error opening output file');
        Continue
      end;

      {$I-}
      BlockWrite(OutFile, Buffer[0], Length(Buffer));
      {$I+}
      if IOResult <> 0 then
        WriteLn('    write error');

      CloseFile(OutFile)
    end;

  SetLength(Buffer, 0);
  SetLength(UncompBuf, 0);
  SetLength(DirEntries, 0)
end;

procedure ScanFiles;
var
  FileName,
  Dir,
  VISName  : string;
  VISFile  : file;
  VISHeader: TVISHeader;
  SearchRec: TSearchRec;
begin
  // read name table if available
  if ParamRec.UseNames then
    ReadNames;

  // get the directory information
  Dir := ExtractFilePath(ParamRec.FileSpec);

  // search for all files matching the file specification
  if FindFirst(ParamRec.FileSpec, faAnyFile - faDirectory,
    SearchRec) = 0 then
  begin
    repeat
      FileName := SearchRec.Name;
      VISName := Dir + FileName;

      WriteLn(VISName);

      if OpenInFile(VISFile, VISName) then
      begin
        if ReadHeader(VISFile, VISHeader) then
          ExtractFiles(VISFile, FileName, VISHeader)
        else
          WriteLn('  not a valid VIS file')
      end
      else
        WriteLn('  not accessible')
    until FindNext(SearchRec) <> 0
  end
  else
    WriteLn(Format('No matches for "%s"', [ParamRec.FileSpec]));

  SysUtils.FindClose(SearchRec)
end;

begin
  try
    Command := ProcessParams;

    if (Command and Command_Info) <> 0 then
      WriteModuleInfo;
    if (Command and Command_Invalid) <> 0 then
      {$IFNDEF CONVERSION_RESTRICTED}
      WriteLn(Format('Usage: %s <VISFiles> [/names=<NameFile> | /generateName' +
        's=<NameFile>] [/unique | /useDefaultName] [/noImgDecrypt] [/override' +
        'Key=<Key>] [/force] [/keyonly] [/BINtoXML | /BINtoVEB | /BINtoVED] [' +
        '/XMLtoVED | /CleanXML] [/info]', [GetModuleNamePrefix]));
      {$ELSE}
      WriteLn(Format('Usage: %s <VISFiles> [/names=<NameFile> | /generateName' +
        's=<NameFile>] [/unique | /useDefaultName] [/noImgDecrypt] [/override' +
        'Key=<Key>] [/force] [/keyonly] [/BINtoXML] [/CleanXML] [/info]',
        [GetModuleNamePrefix]));
      {$ENDIF}
    if (Command and Command_Operate) <> 0 then
    begin
      if not ReadConfig then
        WriteLn(Format('Warning: %s', [LastError]));

      ScanFiles
    end
  except on e: Exception do
    WriteLn('Exception: ', e.Message)
  end
end.
