{ File    : $URL: file:///D:/SVN/svnDelphi/branches/A%20New%20Beginning/VISExt/servicing/src/convert/CleanXML.pas $
  Author  : $Author: Deniz Oezmen $
  Changed : $Date: 2022-05-09 19:13:32 +0200 (Mo, 09 Mai 2022) $
  Revision: $Rev: 626 $

  Visionaire Engine cleanup mechanism for textual XMLs
}
unit CleanXML;

interface

uses
  Shared;

function CleanTextualXML(const XMLBuffer: TBuffer;
  out CleanBuffer: TBuffer): Boolean;

implementation

uses
  MemSearch,
  {$IFNDEF VIS_VERSION_5}
  Names3;
  {$ELSE}
  Names5;
  {$ENDIF}

function CleanInplace(var CleanBuffer: TBuffer;
  NameEndPattern: AnsiString): Boolean;
var
  NameEnd: TBuffer;
  BaseIdx,
  EndIdx,
  OutIdx : Longint;
  ArcDes : AnsiString;
begin
  // The XML file within the game's main VIS archive contains mapping
  // information between a file's original name and its location within the
  // game's archive collection. Since parsing the XML file would be a bit of an
  // overkill here, we're just using a heuristic pattern search, which seems to
  // be working pretty well.

  // set up search patterns
  SetLength(NameEnd, Length(NameEndPattern));
  Move(NameEndPattern[1], NameEnd[0], Length(NameEndPattern));

  // Search the text buffer for the end of names: The terminating combination #"
  // is much more significant than the opening =".
  BaseIdx := 0;
  OutIdx := 0;
  while BaseIdx < High(CleanBuffer) do
  begin
    // search for the end of a name
    EndIdx := SearchMem(SearchMethod_KnuthMorrisPratt, @NameEnd[0],
      @CleanBuffer[BaseIdx], Length(NameEnd), Length(CleanBuffer) - BaseIdx);

    if EndIdx < 0 then
    begin
      // if no match can be found anymore, simply store the rest of the buffer
      EndIdx := High(CleanBuffer) - BaseIdx;
      Move(CleanBuffer[BaseIdx], CleanBuffer[OutIdx], EndIdx + 1);
      Inc(OutIdx, EndIdx + 1)
    end
    else
    begin
      if BaseIdx + EndIdx - ArchiveDesignatorLength + 1 >= 0 then
      begin
        SetLength(ArcDes, ArchiveDesignatorLength);
        Move(CleanBuffer[BaseIdx + EndIdx - ArchiveDesignatorLength + 1],
          ArcDes[1], ArchiveDesignatorLength)
      end
      else
        ArcDes := '';

      if IsArchiveDesignator(ArcDes) then
      begin
        // if this seems to be a valid name, store everything except the archive
        // designator
        Move(CleanBuffer[BaseIdx], CleanBuffer[OutIdx],
          EndIdx - ArchiveDesignatorLength + 1);
        Inc(OutIdx, EndIdx - ArchiveDesignatorLength + 1)
      end
      else
      begin
        // else store everything
        Move(CleanBuffer[BaseIdx], CleanBuffer[OutIdx], EndIdx + 1);
        Inc(OutIdx, EndIdx + 1)
      end
    end;

    BaseIdx := BaseIdx + EndIdx + 1
  end;

  // truncate output data and clean up
  SetLength(CleanBuffer, OutIdx);
  SetLength(NameEnd, 0);

  Result := True
end;

function CleanTextualXML(const XMLBuffer: TBuffer;
  out CleanBuffer: TBuffer): Boolean;
begin
  SetLength(CleanBuffer, Length(XMLBuffer));
  Move(XMLBuffer[0], CleanBuffer[0], Length(XMLBuffer));

  Result := CleanInplace(CleanBuffer, '#"');
  if not Result then
    Exit;

  Result := CleanInplace(CleanBuffer, '#&quot;');
  if not Result then
    Exit;

  Result := CleanInplace(CleanBuffer, '#&apos;')
end;

end.
