{ File    : $URL: file:///D:/SVN/svnDelphi/branches/A%20New%20Beginning/VISExt/servicing/src/utils/CmdLine.pas $
  Author  : $Author: Deniz Oezmen $
  Changed : $Date: 2017-09-27 02:00:28 +0200 (Mi, 27 Sep 2017) $
  Revision: $Rev: 474 $

  Visionaire engine extractor command line handling
}
unit CmdLine;

interface

const
  Command_None    = $00000000;
  Command_Invalid = $00000001;
  Command_Info    = $00000002;
  Command_Operate = $00000004;

type
  TParamRec = record
    FileSpec      : AnsiString;
    NameSpec      : AnsiString;
    GenNames      : Boolean;
    UseNames      : Boolean;
    Unique        : Boolean;
    UseDefaultName: Boolean;
    NoImgDecrypt  : Boolean;
    OverrideKey   : AnsiString;
    Force         : Boolean;
    KeyOnly       : Boolean;
    BINtoXML      : Boolean;
    BINtoVEB      : Boolean;
    BINtoVED      : Boolean;
    XMLtoVED      : Boolean;
    CleanXML      : Boolean;
    Info          : Boolean;
  end;

var
  ParamRec: TParamRec;
  Command : Longword;

function ProcessParams: Longword;
procedure WriteModuleInfo;

implementation

uses
  ConOut,
  ModInfo,
  SysUtils;

function ProcessParams: Longword;
var
  i         : Longint;
  Param,
  ParamLower: AnsiString;
begin
  Result := Command_None;

  with ParamRec do
  begin
    FileSpec       := '';
    NameSpec       := '';
    GenNames       := False;
    UseNames       := False;
    Unique         := False;
    UseDefaultName := False;
    NoImgDecrypt   := False;
    OverrideKey    := '';
    Force          := False;
    KeyOnly        := False;
    BINtoXML       := False;
    BINtoVEB       := False;
    BINtoVED       := False;
    XMLtoVED       := False;
    CleanXML       := False;
    Info           := False;

    // collect information from all parameters
    for i := 1 to ParamCount do
    begin
      Param := ParamStr(i);
      ParamLower := AnsiLowerCase(Param);
      if Copy(ParamLower, 1, 15) = '/generatenames=' then
      begin
        GenNames := True;
        if NameSpec <> '' then
          Result := Result or Command_Invalid;

        NameSpec := Copy(Param, 16, Length(Param) - 15)
      end
      else if Copy(ParamLower, 1, 7) = '/names=' then
      begin
        UseNames := True;
        if NameSpec <> '' then
          Result := Result or Command_Invalid;

        NameSpec := Copy(Param, 8, Length(Param) - 7)
      end
      else if ParamLower = '/unique' then
        Unique := True
      else if ParamLower = '/usedefaultname' then
        UseDefaultName := True
      else if ParamLower = '/noimgdecrypt' then
        NoImgDecrypt := True
      else if Copy(ParamLower, 1, 13) = '/overridekey=' then
      begin
        if OverrideKey <> '' then
          Result := Result or Command_Invalid;

        OverrideKey := Copy(ParamStr(i), 14, Length(ParamStr(i)) - 13)
      end
      else if ParamLower = '/force' then
        Force := True
      else if ParamLower = '/keyonly' then
        KeyOnly := True
      else if ParamLower = '/bintoxml' then
        BINtoXML := True
      {$IFNDEF CONVERSION_RESTRICTED}
      else if ParamLower = '/bintoveb' then
        BINtoVEB := True
      else if ParamLower = '/bintoved' then
        BINtoVED := True
      else if ParamLower = '/xmltoved' then
        XMLtoVED := True
      {$ENDIF}
      else if ParamLower = '/cleanxml' then
        CleanXML := True
      else if ParamLower = '/info' then
        Info := True
      else if Copy(ParamLower, 1, 1) = '/' then
        // any other parameter looking like a switch is unknown
        Result := Result or Command_Invalid
      else if FileSpec = '' then
        FileSpec := Param
      else
        Result := Result or Command_Invalid
    end;

    // check for info flag
    if Info then
      Result := Result or Command_Info;

    // check for invalid parameter combinations
    if GenNames and UseNames then
      Result := Result or Command_Invalid
    else if (GenNames or UseNames) and (NameSpec = '') then
      Result := Result or Command_Invalid
    else if Unique and UseDefaultName then
      Result := Result or Command_Invalid
    else if Byte(BINToXML) + Byte(BINtoVEB) + Byte(BINtoVED) > 1 then
      Result := Result or Command_Invalid;

    // if no invalid parameters have been found, check for minimum operation
    // requirements
    if Result and Command_Invalid = 0 then
    begin
      if FileSpec <> '' then
        Result := Result or Command_Operate
    end;

    // final check: for the command line to be valid, there has to be at least
    // one task to do (be it actual operation or just printing some info)
    if ((Result and (Command_Info or Command_Operate)) = 0) then
      Result := Result or Command_Invalid
  end
end;

procedure WriteModuleInfo;
var
  i         : Longint;
  ModuleInfo: TModuleInfo;
begin
  ModuleInfo := GetModuleInfo;

  WriteLnColored(TextAttr_White, '#');

  for i := Low(ModuleInfo) to High(ModuleInfo) do
    WriteLnColored(TextAttr_White, Format('| %s', [ModuleInfo[i]]));

  WriteLn
end;

end.
